using System;
using System.Threading;
using System.Collections;
using System.Runtime.InteropServices;

namespace Mono.Debugger.Backends
{
	internal class BreakpointManager : IDisposable
	{
		IntPtr _manager;
		Hashtable breakpoints;

		[DllImport("monodebuggerserver")]
		static extern IntPtr mono_debugger_breakpoint_manager_new (BreakpointManagerMutexHandler lock_func, BreakpointManagerMutexHandler unlock_func);

		[DllImport("monodebuggerserver")]
		static extern void mono_debugger_breakpoint_manager_free (IntPtr manager);

		[DllImport("monodebuggerserver")]
		static extern IntPtr mono_debugger_breakpoint_manager_lookup (IntPtr manager, long address);

		[DllImport("monodebuggerserver")]
		static extern int mono_debugger_breakpoint_info_get_id (IntPtr info);

		protected delegate void BreakpointManagerMutexHandler ();
		protected Mutex lock_mutex;

		public BreakpointManager ()
		{
			lock_mutex = new Mutex ();
			breakpoints = new Hashtable ();
			_manager = mono_debugger_breakpoint_manager_new (
				new BreakpointManagerMutexHandler (lock_func),
				new BreakpointManagerMutexHandler (unlock_func));
		}

		void lock_func ()
		{
			lock_mutex.WaitOne ();
		}

		void unlock_func ()
		{
			lock_mutex.ReleaseMutex ();
		}

		internal IntPtr Manager {
			get { return _manager; }
		}

		public Handle LookupBreakpoint (TargetAddress address, out int index)
		{
			lock (this) {
				IntPtr info = mono_debugger_breakpoint_manager_lookup (
					_manager, address.Address);
				if (info == IntPtr.Zero) {
					index = 0;
					return null;
				}

				index = mono_debugger_breakpoint_info_get_id (info);
				return (Handle) breakpoints [index];
			}
		}

		public Handle LookupBreakpoint (int breakpoint)
		{
			lock (this) {
				return (Handle) breakpoints [breakpoint];
			}
		}

		public int InsertBreakpoint (IInferior inferior, Handle handle)
		{
			lock (this) {
				int index = inferior.InsertBreakpoint (handle.Address);
				breakpoints.Add (index, handle);
				return index;
			}
		}

		public void RemoveBreakpoint (IInferior inferior, int index)
		{
			lock (this) {
				inferior.RemoveBreakpoint (index);
				breakpoints.Remove (index);
			}
		}

		public class Handle
		{
			public readonly TargetAddress Address;
			public readonly BreakpointHandle BreakpointHandle;
			public readonly BreakpointCheckHandler CheckHandler;
			public readonly BreakpointHitHandler HitHandler;
			public readonly bool NeedsFrame;
			public readonly object UserData;

			public Handle (TargetAddress address, BreakpointHandle handle,
				       BreakpointCheckHandler check_handler,
				       BreakpointHitHandler hit_handler,
				       bool needs_frame, object user_data)
			{
				this.Address = address;
				this.BreakpointHandle = handle;
				this.CheckHandler = check_handler;
				this.HitHandler = hit_handler;
				this.NeedsFrame = needs_frame;
				this.UserData = user_data;
			}
		}

		//
		// IDisposable
		//

		private bool disposed = false;

		private void check_disposed ()
		{
			if (disposed)
				throw new ObjectDisposedException ("BreakpointManager");
		}

		protected virtual void Dispose (bool disposing)
		{
			// Check to see if Dispose has already been called.
			if (disposed)
				return;

			lock (this) {
				if (disposed)
					return;

				disposed = true;

				mono_debugger_breakpoint_manager_free (_manager);
				_manager = IntPtr.Zero;
			}
		}

		public void Dispose ()
		{
			Dispose (true);
			// Take yourself off the Finalization queue
			GC.SuppressFinalize (this);
		}

		~BreakpointManager ()
		{
			Dispose (false);
		}
	}
}
